@@ -3,6 +3,43 @@
 
 extern struct uwsgi_server uwsgi;
 
+struct uwsgi_route_var *uwsgi_register_route_var(char *name, char *(*func)(struct wsgi_request *, char *, uint16_t, uint16_t *)) {
+
+	struct uwsgi_route_var *old_urv = NULL,*urv = uwsgi.route_vars;
+        while(urv) {
+                if (!strcmp(urv->name, name)) {
+                        return urv;
+                }
+                old_urv = urv;
+                urv = urv->next;
+        }
+
+        urv = uwsgi_calloc(sizeof(struct uwsgi_route_var));
+        urv->name = name;
+	urv->name_len = strlen(name);
+        urv->func = func;
+
+        if (old_urv) {
+                old_urv->next = urv;
+        }
+        else {
+                uwsgi.route_vars = urv;
+        }
+
+        return urv;
+}
+
+struct uwsgi_route_var *uwsgi_get_route_var(char *name, uint16_t name_len) {
+	struct uwsgi_route_var *urv = uwsgi.route_vars;
+	while(urv) {
+		if (!uwsgi_strncmp(urv->name, urv->name_len, name, name_len)) {
+			return urv;
+		}
+		urv = urv->next;
+	}
+	return NULL;
+}
+
 struct uwsgi_buffer *uwsgi_routing_translate(struct wsgi_request *wsgi_req, struct uwsgi_route *ur, char *subject, uint16_t subject_len, char *data, size_t data_len) {
 
 	char *pass1 = data;
@@ -44,9 +81,32 @@ struct uwsgi_buffer *uwsgi_routing_translate(struct wsgi_request *wsgi_req, stru
 			case 2:
 				if (pass1[i] == '}') {
 					uint16_t vallen = 0;
-					char *value = uwsgi_get_var(wsgi_req, key, keylen, &vallen);
+					int need_free = 0;
+					char *value = NULL;
+					char *bracket = memchr(key, '[', keylen);
+					if (bracket && keylen > 0 && key[keylen-1] == ']') {
+						struct uwsgi_route_var *urv = uwsgi_get_route_var(key, bracket - key);
+						if (urv) {
+							need_free = urv->need_free;
+							value = urv->func(wsgi_req, bracket + 1, keylen - (urv->name_len+2), &vallen); 
+						}
+						else {
+							value = uwsgi_get_var(wsgi_req, key, keylen, &vallen);
+						}
+					}
+					else {
+						value = uwsgi_get_var(wsgi_req, key, keylen, &vallen);
+					}
 					if (value) {
-						if (uwsgi_buffer_append(ub, value, vallen)) goto error;
+						if (uwsgi_buffer_append(ub, value, vallen)) {
+							if (need_free) {
+								free(value);
+							}
+							goto error;
+						}
+						if (need_free) {
+							free(value);
+						}
 					}
                                         status = 0;
 					key = NULL;
@@ -108,6 +168,11 @@ int uwsgi_apply_routes_do(struct wsgi_request *wsgi_req, char *subject, uint16_t
 		wsgi_req->route_goto = 0;
 
 		if (!routes->if_func) {
+			// could be a "run"
+			if (!routes->subject) {
+				n = 0;
+				goto run;
+			}
 			if (!subject) {
 				char **subject2 = (char **) (((char *) (wsgi_req)) + routes->subject);
 				uint16_t *subject_len2 = (uint16_t *) (((char *) (wsgi_req)) + routes->subject_len);
@@ -137,6 +202,7 @@ int uwsgi_apply_routes_do(struct wsgi_request *wsgi_req, char *subject, uint16_t
 			}
 		}
 
+run:
 		if (n >= 0) {
 			wsgi_req->is_routing = 1;
 			int ret = routes->func(wsgi_req, routes);
@@ -197,6 +263,8 @@ static void *uwsgi_route_get_condition_func(char *name) {
 
 void uwsgi_opt_add_route(char *opt, char *value, void *foobar) {
 
+	char *space = NULL;
+	char *command = NULL;
 	struct uwsgi_route *old_ur = NULL,*ur = uwsgi.routes;
 	uint64_t pos = 0;
 	while(ur) {
@@ -223,7 +291,12 @@ void uwsgi_opt_add_route(char *opt, char *value, void *foobar) {
 
 	char *route = uwsgi_str(value);
 
-	char *space = strchr(route, ' ');
+	if (!strcmp(foobar, "run")) {
+		command = route;	
+		goto done;
+	}
+
+	space = strchr(route, ' ');
 	if (!space) {
 		uwsgi_log("invalid route syntax\n");
 		exit(1);
@@ -299,7 +372,8 @@ void uwsgi_opt_add_route(char *opt, char *value, void *foobar) {
 		}
 	}
 
-	char *command = space + 1;
+	command = space + 1;
+done:
 	ur->action = uwsgi_str(command);
 
 	char *colon = strchr(command, ':');
@@ -666,6 +740,20 @@ static int uwsgi_route_condition_regexp(struct wsgi_request *wsgi_req, struct uw
         return 0;
 }
 
+static int uwsgi_route_condition_empty(struct wsgi_request *wsgi_req, struct uwsgi_route *ur) {
+
+        struct uwsgi_buffer *ub = uwsgi_routing_translate(wsgi_req, ur, NULL, 0, ur->subject_str, ur->subject_str_len);
+        if (!ub) return -1;
+
+	if (ub->pos == 0) {
+        	uwsgi_buffer_destroy(ub);
+        	return 1;
+	}
+
+        uwsgi_buffer_destroy(ub);
+        return 0;
+}
+
 
 static int uwsgi_route_condition_equal(struct wsgi_request *wsgi_req, struct uwsgi_route *ur) {
 	char *semicolon = memchr(ur->subject_str, ';', ur->subject_str_len);
@@ -806,6 +894,11 @@ void uwsgi_register_embedded_routers() {
         uwsgi_register_route_condition("endswith", uwsgi_route_condition_endswith);
         uwsgi_register_route_condition("regexp", uwsgi_route_condition_regexp);
         uwsgi_register_route_condition("re", uwsgi_route_condition_regexp);
+
+        uwsgi_register_route_condition("empty", uwsgi_route_condition_empty);
+
+        uwsgi_register_route_var("cookie", uwsgi_get_cookie);
+        uwsgi_register_route_var("qs", uwsgi_get_qs);
 }
 
 struct uwsgi_router *uwsgi_register_router(char *name, int (*func) (struct uwsgi_route *, char *)) {
@@ -864,6 +957,9 @@ void uwsgi_routing_dump() {
 		if (routes->label) {
 			uwsgi_log("[rule: %llu] label: %s\n", (unsigned long long ) routes->pos, routes->label);
 		}
+		else if (!routes->subject_str && !routes->if_func) {
+			uwsgi_log("[rule: %llu] action: %s\n", (unsigned long long ) routes->pos, routes->action);
+		}
 		else {
 			uwsgi_log("[rule: %llu] subject: %s %s: %s%s action: %s\n", (unsigned long long ) routes->pos, routes->subject_str, routes->if_func ? "func" : "regexp", routes->if_negate ? "!" : "", routes->regexp, routes->action);
 		}
