@@ -10,9 +10,6 @@
 
 extern struct uwsgi_server uwsgi;
 
-// http status codes list
-extern struct http_status_codes hsc[];
-
 static sapi_module_struct uwsgi_sapi_module;
 
 struct uwsgi_php {
@@ -22,6 +19,9 @@ struct uwsgi_php {
 	struct uwsgi_string_list *index;
 	struct uwsgi_string_list *set;
 	struct uwsgi_string_list *append_config;
+#ifdef UWSGI_PCRE
+	struct uwsgi_regexp_list *app_bypass;
+#endif
 	struct uwsgi_string_list *vars;
 	char *docroot;
 	char *app;
@@ -52,57 +52,23 @@ struct uwsgi_option uwsgi_php_options[] = {
         {"php-allowed-script", required_argument, 0, "list the allowed php scripts (require absolute path)", uwsgi_opt_add_string_list, &uphp.allowed_scripts, 0},
         {"php-server-software", required_argument, 0, "force php SERVER_SOFTWARE", uwsgi_opt_set_str, &uphp.server_software, 0},
         {"php-app", required_argument, 0, "force the php file to run at each request", uwsgi_opt_set_str, &uphp.app, 0},
+        {"php-app-qs", required_argument, 0, "when in app mode force QUERY_STRING to the specified value + REQUEST_URI", uwsgi_opt_set_str, &uphp.app_qs, 0},
+        {"php-fallback", required_argument, 0, "run the specified php script when the request one does not exist", uwsgi_opt_set_str, &uphp.fallback, 0},
+#ifdef UWSGI_PCRE
+        {"php-app-bypass", required_argument, 0, "if the regexp matches the uri the --php-app is bypassed", uwsgi_opt_add_regexp_list, &uphp.app_bypass, 0},
+#endif
         {"php-var", required_argument, 0, "add/overwrite a CGI variable at each request", uwsgi_opt_add_string_list, &uphp.vars, 0},
-        {"php-app-qs", required_argument, 0, "when in app mode force QUERY_STRING to the specified value + PATH_INFO", uwsgi_opt_set_str, &uphp.app_qs, 0},
-	{"php-fallback", required_argument, 0, "run the specified php script when the request one does not exist", uwsgi_opt_set_str, &uphp.fallback, 0},
         {"php-dump-config", no_argument, 0, "dump php config (if modified via --php-set or append options)", uwsgi_opt_true, &uphp.dump_config, 0},
         {0, 0, 0, 0, 0, 0, 0},
 
 };
 
-void uwsgi_php_redirect_to_slash(struct wsgi_request *wsgi_req) {
-
-        struct iovec iov[6];
-
-        wsgi_req->status = 301;
-        iov[0].iov_base = wsgi_req->protocol;
-        iov[0].iov_len = wsgi_req->protocol_len;
-        iov[1].iov_base = " 301 Moved Permanently\r\n";
-        iov[1].iov_len = 24;
-        wsgi_req->headers_size += wsgi_req->socket->proto_writev_header(wsgi_req, iov, 2);
-
-        iov[0].iov_base = "Location: ";
-        iov[0].iov_len = 10;
-        iov[1].iov_base = wsgi_req->path_info;
-        iov[1].iov_len = wsgi_req->path_info_len;
-        iov[2].iov_base = "/";
-        iov[2].iov_len = 1;
-
-        if (wsgi_req->query_string_len > 0) {
-                iov[3].iov_base = "?";
-                iov[3].iov_len = 1;
-                iov[4].iov_base = wsgi_req->query_string;
-                iov[4].iov_len = wsgi_req->query_string_len;
-                iov[5].iov_base = "\r\n\r\n";
-                iov[5].iov_len = 4;
-                wsgi_req->headers_size += wsgi_req->socket->proto_writev_header(wsgi_req, iov, 6);
-                wsgi_req->header_cnt++;
-        }
-        else {
-                iov[3].iov_base = "\r\n\r\n";
-                iov[3].iov_len = 4;
-                wsgi_req->headers_size += wsgi_req->socket->proto_writev_header(wsgi_req, iov, 4);
-                wsgi_req->header_cnt++;
-        }
-}
-
-
 
 static int sapi_uwsgi_ub_write(const char *str, uint str_length TSRMLS_DC)
 {
 	struct wsgi_request *wsgi_req = (struct wsgi_request *) SG(server_context);
 
-	wsgi_req->response_size += wsgi_req->socket->proto_write(wsgi_req, (char *) str, str_length);
+	uwsgi_response_write_body_do(wsgi_req, (char *) str, str_length);
 	if (wsgi_req->write_errors > uwsgi.write_errors_tolerance) {
 		php_handle_aborted_connection();
 		return -1;
@@ -110,147 +76,56 @@ static int sapi_uwsgi_ub_write(const char *str, uint str_length TSRMLS_DC)
 	return str_length;
 }
 
-void uwsgi_php_404(struct wsgi_request *wsgi_req) {
-
-        wsgi_req->status = 404;
-        wsgi_req->headers_size += wsgi_req->socket->proto_write(wsgi_req, "HTTP/1.0 404 Not Found\r\nContent-Type: text/plain\r\n\r\nNot Found", 61);
-}
-
-void uwsgi_php_403(struct wsgi_request *wsgi_req) {
-
-        wsgi_req->status = 403;
-        wsgi_req->headers_size += wsgi_req->socket->proto_write(wsgi_req, "HTTP/1.0 403 Forbidden\r\nContent-Type: text/plain\r\n\r\nForbidden", 61);
-
-}
-
-
 static int sapi_uwsgi_send_headers(sapi_headers_struct *sapi_headers)
 {
 	sapi_header_struct *h;
 	zend_llist_position pos;
-	struct iovec iov[6];
-	char status[4];
-	struct http_status_codes *http_sc;
 
 	if (SG(request_info).no_headers == 1) {
                 return SAPI_HEADER_SENT_SUCCESSFULLY;
         }
 
 	struct wsgi_request *wsgi_req = (struct wsgi_request *) SG(server_context);
-	wsgi_req->status = SG(sapi_headers).http_response_code;
-	if (!wsgi_req->status) wsgi_req->status = 200;
 
 	if (!SG(sapi_headers).http_status_line) {
-
-		iov[0].iov_base = wsgi_req->protocol;
-		iov[0].iov_len = wsgi_req->protocol_len;
-
-
-		iov[1].iov_base = " ";
-		iov[1].iov_len = 1;
-
-		uwsgi_num2str2n(wsgi_req->status, status, 4);
-
-		iov[2].iov_base = status;
-		iov[2].iov_len = 3;
-
-		iov[3].iov_base = " ";
-		iov[3].iov_len = 1;
-
-		// get the status code
-        	for (http_sc = hsc; http_sc->message != NULL; http_sc++) {
-                	if (!strncmp(http_sc->key, status, 3)) {
-                        	iov[4].iov_base = (char *) http_sc->message;
-                        	iov[4].iov_len = http_sc->message_size;
-                        	break;
-                	}
-        	}
-
-        	if (iov[4].iov_len == 0) {
-                	iov[4].iov_base = "Unknown";
-                	iov[4].iov_len =  7;
-        	}
-
-		iov[5].iov_base = "\r\n";
-		iov[5].iov_len = 2;
-
-		wsgi_req->headers_size += wsgi_req->socket->proto_writev_header(wsgi_req, iov, 6);
+		char status[4];
+		int hrc = SG(sapi_headers).http_response_code;
+		if (!hrc) hrc = 200;
+		uwsgi_num2str2n(hrc, status, 4);
+		uwsgi_response_prepare_headers(wsgi_req, status, 3);
 	}
 	else {
-		iov[0].iov_base = SG(sapi_headers).http_status_line;
-		iov[0].iov_len = strlen(iov[0].iov_base);
-		iov[1].iov_base = "\r\n";
-		iov[1].iov_len = 2;
-		wsgi_req->headers_size += wsgi_req->socket->proto_writev_header(wsgi_req, iov, 2);
+		char *sl = SG(sapi_headers).http_status_line;
+		uwsgi_response_prepare_headers(wsgi_req, sl, strlen(sl));
 	}
 	
 	h = zend_llist_get_first_ex(&sapi_headers->headers, &pos);
 	while (h) {
-		iov[0].iov_base = h->header;
-		iov[0].iov_len = h->header_len;
-		iov[1].iov_base = "\r\n";
-		iov[1].iov_len = 2;
-
-		wsgi_req->headers_size += wsgi_req->socket->proto_writev_header(wsgi_req, iov, 2);	
-		wsgi_req->header_cnt++;
+		uwsgi_response_add_header(wsgi_req, NULL, 0, h->header, h->header_len);
 		h = zend_llist_get_next_ex(&sapi_headers->headers, &pos);
 	}
 
-	struct uwsgi_string_list *ah = uwsgi.additional_headers;
-        while(ah) {
-                        iov[0].iov_base = ah->value;
-                        iov[0].iov_len = ah->len;
-                        iov[1].iov_base = "\r\n";
-                        iov[1].iov_len = 2;
-                        wsgi_req->headers_size += wsgi_req->socket->proto_writev_header(wsgi_req, iov, 2);
-                        wsgi_req->header_cnt++;
-                        ah = ah->next;
-        }
-
-
-	wsgi_req->headers_size += wsgi_req->socket->proto_write_header(wsgi_req, "\r\n", 2);
-	
 	return SAPI_HEADER_SENT_SUCCESSFULLY;
 }
 
 static int sapi_uwsgi_read_post(char *buffer, uint count_bytes TSRMLS_DC)
 {
 	uint read_bytes = 0;
-	ssize_t len;
-	int fd = -1;
 	
 	struct wsgi_request *wsgi_req = (struct wsgi_request *) SG(server_context);
 
-	if (wsgi_req->body_as_file) {
-                fd = fileno((FILE *)wsgi_req->async_post);
-        }
-        else if (uwsgi.post_buffering > 0) {
-                if (wsgi_req->post_cl > (size_t) uwsgi.post_buffering) {
-                        fd = fileno((FILE *)wsgi_req->async_post);
-                }
-        }
-        else {
-                fd = wsgi_req->poll.fd;
-        }
-
-
         count_bytes = MIN(count_bytes, wsgi_req->post_cl - SG(read_post_bytes));
 
-	// data in memory
-	if (fd == -1) {
-		if (count_bytes > 0) {
-			memcpy(buffer, wsgi_req->post_buffering_buf + wsgi_req->post_pos, count_bytes);
-			wsgi_req->post_pos += count_bytes;
-		}
-		return count_bytes;
-	}
-
         while (read_bytes < count_bytes) {
-                len = read(fd, buffer + read_bytes, count_bytes - read_bytes);
-		if (len <= 0) {
-			break;
+		ssize_t rlen = 0;
+		char *buf = uwsgi_request_body_read(wsgi_req, count_bytes - read_bytes, &rlen);
+		if (buf == uwsgi.empty) break;
+		if (buf) {
+			memcpy(buffer, buf, rlen);
+			read_bytes += rlen;
+			continue;
 		}
-                read_bytes += len;
+		break;
         }
 
         return read_bytes;
@@ -332,7 +207,7 @@ static sapi_module_struct uwsgi_sapi_module;
 
 
 void uwsgi_php_append_config(char *filename) {
-	int file_size = 0;
+	size_t file_size = 0;
         char *file_content = uwsgi_open_and_read(filename, &file_size, 1, NULL);
 	uwsgi_sapi_module.ini_entries = realloc(uwsgi_sapi_module.ini_entries, uphp.ini_size + file_size);
 	memcpy(uwsgi_sapi_module.ini_entries + uphp.ini_size, file_content, file_size);
@@ -370,25 +245,53 @@ PHP_FUNCTION(uwsgi_masterpid) {
 	RETURN_LONG(0);
 }
 
+PHP_FUNCTION(uwsgi_cache_exists) {
+
+        char *key = NULL;
+        int keylen = 0;
+        char *cache = NULL;
+
+        if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s|s", &key, &keylen, &cache) == FAILURE) {
+                RETURN_NULL();
+        }
+
+        if (uwsgi_cache_magic_exists(key, keylen, cache)) {
+                RETURN_TRUE;
+        }
+
+        RETURN_NULL();
+}
+
+PHP_FUNCTION(uwsgi_cache_clear) {
+
+        char *cache = NULL;
+
+        if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "|s", &cache) == FAILURE) {
+                RETURN_NULL();
+        }
+
+        if (!uwsgi_cache_magic_clear(cache)) {
+                RETURN_TRUE;
+        }
+
+        RETURN_NULL();
+}
+
+
 PHP_FUNCTION(uwsgi_cache_del) {
 	
 	char *key = NULL;
         int keylen = 0;
+	char *cache = NULL;
 
-        if (!uwsgi.cache_max_items)
-                RETURN_NULL();
-
-        if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s", &key, &keylen) == FAILURE) {
+        if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s|s", &key, &keylen, &cache) == FAILURE) {
                 RETURN_NULL();
         }
 
-	uwsgi_wlock(uwsgi.cache_lock);
-        if (uwsgi_cache_del(key, keylen, 0)) {
-                uwsgi_rwunlock(uwsgi.cache_lock);
+        if (!uwsgi_cache_magic_del(key, keylen, cache)) {
 		RETURN_TRUE;
         }
 
-        uwsgi_rwunlock(uwsgi.cache_lock);
 	RETURN_NULL();
 }
 
@@ -396,23 +299,22 @@ PHP_FUNCTION(uwsgi_cache_get) {
 
 	char *key = NULL;
 	int keylen = 0;
+	char *cache = NULL;
 	uint64_t valsize;
 
-	if (!uwsgi.cache_max_items)
+	if (!uwsgi.caches)
 		RETURN_NULL();
 
-	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s", &key, &keylen) == FAILURE) {
+	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s|s", &key, &keylen, &cache) == FAILURE) {
                 RETURN_NULL();
         }
 
-	uwsgi_rlock(uwsgi.cache_lock);
-	char *value = uwsgi_cache_get(key, keylen, &valsize);
+	char *value = uwsgi_cache_magic_get(key, keylen, &valsize, cache);
 	if (value) {
 		char *ret = estrndup(value, valsize);
-		uwsgi_rwunlock(uwsgi.cache_lock);
+		free(value);
 		RETURN_STRING(ret, 0);
 	}
-	uwsgi_rwunlock(uwsgi.cache_lock);
 	RETURN_NULL();
 }
 
@@ -422,24 +324,18 @@ PHP_FUNCTION(uwsgi_cache_set) {
 	char *value = NULL;
 	int vallen;
 	uint64_t expires = 0;
+	char *cache = NULL;
 
-	if (!uwsgi.cache_max_items)
+	if (!uwsgi.caches)
 		RETURN_NULL();
 
-	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ss|l", &key, &keylen, &value, &vallen, &expires) == FAILURE) {
+	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ss|ls", &key, &keylen, &value, &vallen, &expires, &cache) == FAILURE) {
                 RETURN_NULL();
         }
 
-	if ((uint64_t)vallen > uwsgi.cache_blocksize) {
-		RETURN_NULL();
-	}
-	
-	uwsgi_wlock(uwsgi.cache_lock);
-	if (uwsgi_cache_set(key, keylen, value, vallen, expires, 0)) {
-		uwsgi_rwunlock(uwsgi.cache_lock);
+	if (uwsgi_cache_magic_set(key, keylen, value, vallen, expires, 0, cache)) {
 		RETURN_TRUE;
 	}
-	uwsgi_rwunlock(uwsgi.cache_lock);
 	RETURN_NULL();
 	
 }
@@ -450,24 +346,18 @@ PHP_FUNCTION(uwsgi_cache_update) {
         char *value = NULL;
         int vallen;
         uint64_t expires = 0;
+        char *cache = NULL;
 
-	if (!uwsgi.cache_max_items)
-		RETURN_NULL();
+        if (!uwsgi.caches)
+                RETURN_NULL();
 
-        if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ss|l", &key, &keylen, &value, &vallen, &expires) == FAILURE) {
+        if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ss|ls", &key, &keylen, &value, &vallen, &expires, &cache) == FAILURE) {
                 RETURN_NULL();
         }
 
-        if ((uint64_t)vallen > uwsgi.cache_blocksize) {
-                RETURN_NULL();
-        }
-
-        uwsgi_wlock(uwsgi.cache_lock);
-        if (uwsgi_cache_set(key, keylen, value, vallen, expires, UWSGI_CACHE_FLAG_UPDATE)) {
-                uwsgi_rwunlock(uwsgi.cache_lock);
+        if (uwsgi_cache_magic_set(key, keylen, value, vallen, expires, UWSGI_CACHE_FLAG_UPDATE, cache)) {
                 RETURN_TRUE;
         }
-        uwsgi_rwunlock(uwsgi.cache_lock);
         RETURN_NULL();
 
 }
@@ -521,14 +411,12 @@ PHP_FUNCTION(uwsgi_rpc) {
 
 	// response must always be freed
         char *response = uwsgi_do_rpc(node, func, num_args - 2, argv, argvs, &size);
-
-        if (size > 0) {
+        if (response) {
 		// here we do not free varargs for performance reasons
 		char *ret = estrndup(response, size);
 		free(response);
 		RETURN_STRING(ret, 0);
         }
-	free(response);
 
 clear:
 	efree(varargs);
@@ -572,18 +460,22 @@ zend_function_entry uwsgi_php_functions[] = {
 	PHP_FE(uwsgi_worker_id,   NULL)
 	PHP_FE(uwsgi_masterpid,   NULL)
 	PHP_FE(uwsgi_signal,   NULL)
+
 	PHP_FE(uwsgi_rpc,   NULL)
+
 	PHP_FE(uwsgi_cache_get,   NULL)
 	PHP_FE(uwsgi_cache_set,   NULL)
 	PHP_FE(uwsgi_cache_update,   NULL)
 	PHP_FE(uwsgi_cache_del,   NULL)
+	PHP_FE(uwsgi_cache_clear,   NULL)
+	PHP_FE(uwsgi_cache_exists,   NULL)
 	{ NULL, NULL, NULL},
 };
 
 PHP_MINFO_FUNCTION(uwsgi_php_minfo) {
 	php_info_print_table_start( );
 	php_info_print_table_row(2, "uwsgi api", "enabled");
-	if (uwsgi.cache_max_items > 0) {
+	if (uwsgi.caches) {
 		php_info_print_table_row(2, "uwsgi cache", "enabled");
 	}
 	else {
@@ -654,8 +546,6 @@ static sapi_module_struct uwsgi_sapi_module = {
 
 int uwsgi_php_init(void) {
 
-	struct http_status_codes *http_sc;
-
 	struct uwsgi_string_list *pset = uphp.set;
 	struct uwsgi_string_list *append_config = uphp.append_config;
 
@@ -689,11 +579,6 @@ int uwsgi_php_init(void) {
 
 	uwsgi_sapi_module.startup(&uwsgi_sapi_module);
 
-	// filling http status codes
-        for (http_sc = hsc; http_sc->message != NULL; http_sc++) {
-                http_sc->message_size = strlen(http_sc->message);
-        }
-
 	uwsgi_log("PHP %s initialized\n", PHP_VERSION);
 
 	return 0;
@@ -761,6 +646,7 @@ int uwsgi_php_request(struct wsgi_request *wsgi_req) {
 	char *path_info = NULL;
 	size_t real_filename_len = 0;
 	struct stat php_stat;
+	char *filename = NULL;
 
 	zend_file_handle file_handle;
 
@@ -792,6 +678,16 @@ int uwsgi_php_request(struct wsgi_request *wsgi_req) {
 	wsgi_req->document_root_len = strlen(wsgi_req->document_root);
 
 	if (uphp.app) {
+#ifdef UWSGI_PCRE
+		struct uwsgi_regexp_list *bypass = uphp.app_bypass;
+		while (bypass) {
+                        if (uwsgi_regexp_match(bypass->pattern, bypass->pattern_extra, wsgi_req->uri, wsgi_req->uri_len) >= 0) {
+				goto oldstyle;
+                        }
+                        bypass = bypass->next;
+                }
+#endif
+
 		strcpy(real_filename, uphp.app);	
 		if (wsgi_req->path_info_len == 1 && wsgi_req->path_info[0] == '/') {
 			goto appready;
@@ -819,7 +715,11 @@ appready:
 		goto secure2;
 	}
 
-	char *filename = uwsgi_concat4n(wsgi_req->document_root, wsgi_req->document_root_len, "/", 1, wsgi_req->path_info, wsgi_req->path_info_len, "", 0);
+#ifdef UWSGI_PCRE
+oldstyle:
+#endif
+
+	filename = uwsgi_concat4n(wsgi_req->document_root, wsgi_req->document_root_len, "/", 1, wsgi_req->path_info, wsgi_req->path_info_len, "", 0);
 
 	if (uwsgi_php_walk(wsgi_req, filename, wsgi_req->document_root, wsgi_req->document_root_len, &path_info)) {
 		free(filename);
@@ -827,7 +727,7 @@ appready:
 			filename = uwsgi_str(uphp.fallback);
 		}
 		else {
-			uwsgi_php_404(wsgi_req);
+			uwsgi_404(wsgi_req);
 			return -1;
 		}
 	}
@@ -848,7 +748,7 @@ appready:
 
 	if (!realpath(filename, real_filename)) {
 		free(filename);
-		uwsgi_php_404(wsgi_req);
+		uwsgi_404(wsgi_req);
 		return -1;
 	}
 
@@ -863,7 +763,7 @@ appready:
 			}
 			usl = usl->next;
 		}
-		uwsgi_php_403(wsgi_req);
+		uwsgi_403(wsgi_req);
 		uwsgi_log("PHP security error: %s is not under an allowed docroot\n", real_filename);
 		return -1;
 	}
@@ -871,7 +771,7 @@ appready:
 secure:
 
 	if (stat(real_filename, &php_stat)) {
-                uwsgi_php_404(wsgi_req);
+                uwsgi_404(wsgi_req);
                 return UWSGI_OK;
         }
 
@@ -881,7 +781,7 @@ secure:
                 if (orig_path_info_len == 0 || (orig_path_info_len > 0 && orig_path_info[orig_path_info_len-1] != '/')) {
 			wsgi_req->path_info = orig_path_info;
 			wsgi_req->path_info_len = orig_path_info_len;
-                        uwsgi_php_redirect_to_slash(wsgi_req);
+                        uwsgi_redirect_to_slash(wsgi_req);
                         return UWSGI_OK;
                 }
                 struct uwsgi_string_list *upi = uphp.index;
@@ -902,7 +802,7 @@ secure:
                 }
 
                 if (!found) {
-                        uwsgi_php_404(wsgi_req);
+                        uwsgi_404(wsgi_req);
                         return UWSGI_OK;
                 }
 
@@ -921,7 +821,7 @@ secure:
 			}
                         usl = usl->next;
                 }
-                uwsgi_php_403(wsgi_req);
+                uwsgi_403(wsgi_req);
                 uwsgi_log("PHP security error: %s does not end with an allowed extension\n", real_filename);
                 return -1;
 	}
@@ -939,7 +839,7 @@ secure2:
                         }
                         usl = usl->next;
                 }
-                uwsgi_php_403(wsgi_req);
+                uwsgi_403(wsgi_req);
                 uwsgi_log("PHP security error: %s is not an allowed script\n", real_filename);
                 return -1;
         }
@@ -984,7 +884,7 @@ secure3:
         file_handle.opened_path = NULL;
 
         if (php_request_startup(TSRMLS_C) == FAILURE) {
-		internal_server_error(wsgi_req, "unable to start php request");
+		uwsgi_500(wsgi_req);
                 return -1;
         }
 
