@@ -1106,6 +1106,14 @@ struct uwsgi_route_condition {
 	struct uwsgi_route_condition *next;
 };
 
+struct uwsgi_route_var {
+	char *name;
+	uint16_t name_len;
+	char *(*func)(struct wsgi_request *, char *, uint16_t, uint16_t *);
+	int need_free;
+	struct uwsgi_route_var *next;
+};
+
 struct uwsgi_router {
 	char *name;
 	int (*func) (struct uwsgi_route *, char *);
@@ -1236,6 +1244,9 @@ struct uwsgi_router {
 	char *referer;
 	uint16_t referer_len;
 
+	char *cookie;
+	uint16_t cookie_len;
+
 	char *path_info;
 	uint16_t path_info_len;
 	int path_info_pos;
@@ -1383,6 +1394,11 @@ struct uwsgi_router {
 	// internal routing goto instruction
 	uint32_t route_goto;
 
+	struct uwsgi_buffer *cache_it;
+	struct uwsgi_buffer *cache_it_to;
+	uint64_t cache_it_expires;
+	struct uwsgi_buffer *cached_response;
+
 	struct msghdr msg;
 	union {
 		struct cmsghdr cmsg;
@@ -2068,6 +2084,7 @@ struct uwsgi_server {
 	struct uwsgi_router *routers;
 	struct uwsgi_route *routes;
 	struct uwsgi_route_condition *route_conditions;
+	struct uwsgi_route_var *route_vars;
 #endif
 
 	int single_interpreter;
@@ -3585,6 +3602,7 @@ int uwsgi_buffer_insert_chunked(struct uwsgi_buffer *, size_t, size_t);
 int uwsgi_buffer_append_chunked(struct uwsgi_buffer *, size_t);
 int uwsgi_buffer_append_json(struct uwsgi_buffer *, char *, size_t);
 int uwsgi_buffer_set_uh(struct uwsgi_buffer *, uint8_t, uint8_t);
+struct uwsgi_buffer *uwsgi_buffer_from_file(char *);
 
 ssize_t uwsgi_buffer_write_simple(struct wsgi_request *, struct uwsgi_buffer *);
 
@@ -3904,6 +3922,12 @@ char *uwsgi_deflate(z_stream *, char *, size_t, size_t *);
 void uwsgi_crc32(uint32_t *, char *, size_t);
 #endif
 
+char *uwsgi_get_cookie(struct wsgi_request *, char *, uint16_t, uint16_t *);
+char *uwsgi_get_qs(struct wsgi_request *, char *, uint16_t, uint16_t *);
+
+struct uwsgi_route_var *uwsgi_get_route_var(char *, uint16_t);
+struct uwsgi_route_var *uwsgi_register_route_var(char *, char *(*)(struct wsgi_request *, char *, uint16_t, uint16_t *));
+
 void uwsgi_check_emperor(void);
 #ifdef UWSGI_AS_SHARED_LIBRARY
 int uwsgi_init(int, char **, char **);
